#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2015-2016 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr


from collections import deque
from itertools import combinations
import cv2
import networkx as nx
from pyannote.core import Annotation
from pyannote.core.time import _t_iter as getLabelGenerator
from tqdm import tqdm
import warnings

try:
    # Python 3
    from functools import lru_cache
except Exception as e:
    # Python 2
    from ..utils.lru_cache import lru_cache

def product_lookahead(iterable, lookahead):

    cache = deque([], lookahead + 1)
    for item in iterable:
        cache.append(item)

        if len(cache) < lookahead + 1:
            continue

        for j in range(lookahead):
            yield cache[0], cache[j+1]

    cache.popleft()
    for item1, item2 in combinations(cache, 2):
        yield item1, item2


class Thread(object):
    """Shot threading based on ORB features

    Can also be used as post-processing to clean up a (possibly over-segmented)
    segmentation into shots.

    """
    def __init__(self, video, shot=None, min_match=20, lookahead=24,
                 verbose=False):
        """
        Parameters
        ----------
        shot : iterable, optional
            Segment iterator.

        """
        super(Thread, self).__init__()

        self.video = video

        self.lookahead = lookahead
        if shot is None:
            shot = Shot(video)
        self.shot = shot

        self.verbose = verbose

        # ORB (non-patented SIFT alternative) extraction
        self._orb = cv2.ORB()

        # # brute-force ORB matching
        # self._bfmatcher = cv2.BFMatcher(cv2.NORM_HAMMING, crossCheck=False)

        # fast approximate nearest neighbord
        FLANN_INDEX_LSH = 6
        index_params = dict(algorithm=FLANN_INDEX_LSH,
                            table_number=6,
                            key_size=12,
                            multi_probe_level=1)
        search_params = dict(checks=50)
        self._flann = cv2.FlannBasedMatcher(index_params, search_params)

        self.min_match = min_match


    # _threads_graph method calls _compute_orb with the same "t" over and over.
    # we cache "maxsize" last calls to avoid recomputing ORB features
    @lru_cache(maxsize=128, typed=False)
    def _compute_orb(self, t):
        try:
            rgb = self.video(t)
            gray = cv2.cvtColor(rgb, cv2.COLOR_RGB2GRAY)
            _, descriptors = self._orb.detectAndCompute(gray, None)

        except IOError as e:
            warnings.warn("unable to reach t = {t:.3f}".format(t=t))
            descriptors = None

        return descriptors

    def _match(self, orb1, orb2):
        """Check whether there is a match between orb1 and orb2"""

        if orb1 is None or orb2 is None:
            return False

        # matches = self._bfmatcher.match(orb1, orb2)
        matches = self._flann.knnMatch(orb1, orb2, k=2)

        count = 0

        for twoNearestNeighbors in matches:
            if len(twoNearestNeighbors) < 2:
                continue
            best, secondBest = twoNearestNeighbors
            if best.distance < 0.7 * secondBest.distance:
                count = count + 1

        return count > self.min_match

    def _threads_graph(self):

        # 5-frames collar
        collar = 5. / self.video._fps

        # build threading graph by comparing each shot
        # to 'lookahead' following shots
        threads = nx.Graph()

        generator = product_lookahead(self.shot, self.lookahead)
        if self.verbose:
            generator = tqdm(iterable=generator,
                             total=len(self.shot) * self.lookahead,
                             leave=True, mininterval=1.,
                             unit='shot pairs', unit_scale=True)

        for current, following in generator:
            orbLast = self._compute_orb(current.end - collar)
            orbFirst = self._compute_orb(following.start + collar)
            threads.add_node(current)
            if self._match(orbLast, orbFirst):
                threads.add_edge(current, following)
        threads.add_node(following)
        return threads

    def __call__(self):

        # list of chronologically sorted list of shots
        graph = self._threads_graph()
        threads = [sorted(cc) for cc in nx.connected_components(graph)]

        annotation = Annotation()
        labelGenerator = getLabelGenerator()

        # chronologically sorted threads (based on their first shot)
        for thread in sorted(threads, key=lambda thread: thread[0]):
            label = next(labelGenerator)
            for shot in thread:
                annotation[shot] = label

        return annotation.smooth()



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2016 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

"""Tracking by detection"""

import itertools
import numpy as np
import networkx as nx
from munkres import Munkres
import dlib

FORWARD = 'forward'
BACKWARD = 'backward'
DETECTION = 'detection'
ERROR = 'error'


def get_segment_generator(segmentation):
    """Time-driven segment generator"""
    t = yield
    for segment in segmentation:
        T = segment.end

        while True:
            # loop until a large enough t is sent to the generator
            if T > t:
                t = yield
                continue

            # else, we found a new segment
            t = yield T
            break


def get_min_max_t(track):
    """Get track stat and end times"""
    m = min(t for t, _, _ in track)
    M = max(t for t, _, _ in track)
    return (m, M)


class TrackingByDetection(object):
    """(Forward/backward) tracking by detection

    Parameters
    ----------
    detect_func : func
        Detection function. Should take video frame as input and return list
        (or iterable) of detections as (left, top, right, bottom) tuples.
    min_confidence : float, optional
        Kill trackers whose confidence goes below this value. Defaults to 10.
    min_overlap_ratio : float, optional
        Do not associate trackers and detections if their overlap ratio goes
        below this value. Defaults to 0.3.
    max_gap : float, optional
        Bridge gaps with duration shorter than this value.

    Usage
    -----
    >>> from pyannote.video import Video, TrackingByDetection
    >>> video = Video(path_to_video)
    >>> # load segmentation into shots
    >>> tracking = TrackingByDetection()
    >>> for face_track in tracking(video, shots):
    ...     # do something with face track
    ...     pass
    """

    def __init__(self, detect_func,
                 min_confidence=10., min_overlap_ratio=0.3, max_gap=0.):

        super(TrackingByDetection, self).__init__()

        self.detect_func = detect_func
        self.min_confidence = min_confidence
        self.min_overlap_ratio = min_overlap_ratio
        self.max_gap = max_gap

        self._hungarian = Munkres()

    def _kill_tracker(self, identifier):
        """Kill specific tracker"""
        del self._trackers[identifier]
        del self._confidences[identifier]
        del self._previous[identifier]

    def _match(self, rectangle1, rectangle2):
        overlap = rectangle1.intersect(rectangle2).area()
        if ((overlap < self.min_overlap_ratio * rectangle1.area()) or
            (overlap < self.min_overlap_ratio * rectangle2.area())):
            overlap = 0.
        return overlap

    def _associate(self, trackers, detections):
        """Associate trackers and detections with Hungarian algorithm

        Parameters
        ----------
        trackers : dict
            Dictionary where values are current trackers
            and keys are trackers identifiers.
        detections : list
            List of detections

        Returns
        -------
        match : dict
            Dictionary where values are trackers
            and keys are matched detection indices.
        """

        n_trackers, n_detections = len(trackers), len(detections)

        if n_trackers < 1 or n_detections < 1:
            return dict()

        n = max(n_trackers, n_detections)
        overlap_area = np.zeros((n, n))

        # list of (identifier, tracker) tuple
        trackers_ = trackers.items()
        for t, (identifier, tracker) in enumerate(trackers_):
            position = tracker.get_position()
            for d, detection in enumerate(detections):
                rectangle = dlib.drectangle(*detection)
                overlap_area[t, d] = self._match(position, rectangle)

        # find the best one-to-one mapping
        match = {}
        mapping = self._hungarian.compute(np.max(overlap_area) - overlap_area)
        for t, d in mapping:

            if t >= n_trackers or d >= n_detections:
                continue

            if overlap_area[t, d] > 0.:
                identifier, _ = trackers_[t]
                match[d] = identifier

        return match

    def _track(self, direction=FORWARD):
        """Actual tracking based on existing detections"""

        if direction == FORWARD:
            frame_cache = self._frame_cache
        elif direction == BACKWARD:
            frame_cache = reversed(self._frame_cache)
        else:
            raise NotImplementedError()

        self._trackers = {}
        self._confidences = {}
        self._previous = {}
        new_identifier = 0

        for t, frame in frame_cache:

            # update trackers & end those with low confidence
            for identifier, tracker in list(self._trackers.items()):
                confidence = tracker.update(frame)
                self._confidences[identifier] = confidence
                if confidence < self.min_confidence:
                    self._kill_tracker(identifier)

            # match trackers with detections at time t
            detections = [d for _, d, status in self._tracking_graph[t]
                          if status == DETECTION]
            match = self._associate(self._trackers, detections)

            # process all matched trackers
            for d, identifier in match.items():

                # connect the previous position of the tracker
                # to the (current) associated detection
                current = (t, detections[d], DETECTION)
                self._tracking_graph.add_edge(
                    self._previous[identifier], current,
                    confidence=self._confidences[identifier])

                # end the tracker
                self._kill_tracker(identifier)

            # process all unmatched trackers
            for identifier, tracker in self._trackers.items():

                # connect the previous position of the tracker
                # to the current position of the tracker
                position = tracker.get_position()
                position = (
                    int(round(position.left())),
                    int(round(position.top())),
                    int(round(position.right())),
                    int(round(position.bottom()))
                )
                current = (t, position, direction)
                self._tracking_graph.add_edge(
                    self._previous[identifier], current,
                    confidence=self._confidences[identifier])

                # save current position of the tracker for next iteration
                self._previous[identifier] = current

            # start new trackers for all detections
            for d, detection in enumerate(detections):

                # start new tracker
                new_tracker = dlib.correlation_tracker()
                new_tracker.start_track(frame, dlib.drectangle(*detection))
                self._trackers[new_identifier] = new_tracker

                # save previous (t, position, status) tuple
                current = (t, detection, DETECTION)
                self._previous[new_identifier] = current

                # increment tracker identifier
                new_identifier = new_identifier + 1

    def _fix(self, track):
        """Fix track by merging matching forward/backward tracklets"""

        fixed_track = []
        for t, group in itertools.groupby(sorted(track), key=lambda x: x[0]):

            group = list(group)

            # make sure all positions are overlap enough
            error = False
            for (_, pos1, _), (_, pos2, _) in itertools.combinations(group, 2):

                rectangle1 = dlib.drectangle(*pos1)
                rectangle2 = dlib.drectangle(*pos2)

                if self._match(rectangle1, rectangle2) == 0:
                    error = True
                    break

            # status
            status = "+".join(
                sorted((status for _, _, status in group),
                       key=lambda s: {DETECTION: 2,
                                      FORWARD: 1,
                                      BACKWARD: 3}[s]))
            if error:
                status = "error({0})".format(status)

            # average position
            pos = tuple(int(round(v))
                        for v in np.mean(np.vstack([p for _, p, _ in group]),
                                         axis=0))

            fixed_track.append((t, pos, status))

        return fixed_track

    def _fill_gaps(self, tracks):

        # sort tracks by start and end timestamps
        tracks = sorted(tracks, key=get_min_max_t)

        # build graph where nodes are tracks and where matching tracks
        # less than "max_gap" away are connected
        graph = nx.Graph()
        for i in xrange(len(tracks)):
            graph.add_node(i)

        for i, j in itertools.combinations(xrange(len(tracks)), 2):

            # only try to match tracks with a short gap between them
            ti = tracks[i][-1][0]
            tj = tracks[j][0][0]
            if (tj < ti) or (tj - ti > self.max_gap):
                continue

            # match tracks whose last and first position match
            rectangle1 = dlib.drectangle(*tracks[i][-1][1])
            rectangle2 = dlib.drectangle(*tracks[j][0][1])
            if self._match(rectangle1, rectangle2):
                graph.add_edge(i, j)

        # merge tracks that are in the same connected component
        merged_tracks = []
        for group in nx.connected_components(graph):
            track = [item for t in sorted(group) for item in tracks[t]]
            merged_tracks.append(track)

        return merged_tracks

    def _forward_backward(self):

        # forward tracking
        self._track(direction=FORWARD)

        # backward tracking
        self._track(direction=BACKWARD)

        # remove timestamps
        timestamps = [t for t in self._tracking_graph
                      if not isinstance(t, tuple)]

        self._tracking_graph.remove_nodes_from(timestamps)

        # tracks are connected components in tracking graph
        tracks = nx.connected_components(
            self._tracking_graph.to_undirected(reciprocal=False))

        # merge matching backward/forward tracks
        tracks = [self._fix(track) for track in tracks]

        # fill gaps
        tracks = self._fill_gaps(tracks)

        # sort tracks by start and end timestamps
        for track in sorted(tracks, key=get_min_max_t):
            yield track

    def _reset(self):
        """Reset tracking"""
        self._frame_cache = []
        self._tracking_graph = nx.DiGraph()

    def __call__(self, video, segmentation):
        """
        Parameters
        ----------
        video : Video
        segmentation :
        """

        segment_generator = get_segment_generator(segmentation)
        segment_generator.send(None)
        self._reset()

        for t, frame in video:

            segment = segment_generator.send(t)

            if segment:

                # forward/backward tracking
                for track in self._forward_backward():
                    yield track

                # start fresh for next segment
                self._reset()

            # cache frame (for faster tracking)
            self._frame_cache.append((t, frame))

            # detection
            self._tracking_graph.add_node(t)
            for detection in self.detect_func(frame):
                self._tracking_graph.add_edge(t, (t, detection, DETECTION))

        for track in self._forward_backward():
            yield track



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2015-2016 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr


def find_cut_edge(graph):

    edges = nx.minimum_edge_cut(graph)
    if len(edges) != 1:
        return

    cut_edge = edges.pop()
    yield cut_edge

    # remove cut edge and process resulting (2) subgraphs
    graph.remove_edge(*cut_edge)
    for subgraph in nx.connected_component_subgraphs(graph):
        for cut_edge in find_cut_edge(subgraph):
            yield cut_edge


class Scene(object):
    """Segmentation into scenes"""

    def __init__(self, video, thread=None):
        super(Scene, self).__init__()
        self.video = video
        if thread is None:
            thread = Thread(video)()



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2015-2016 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

"""Face detection and tracking

The standard pipeline is the following (with optional face tracking)

face detection => (face tracking =>) landmarks detection => feature extraction

Usage:
  pyannote-face track [options] [--verbose] <video> <shot.json> <output>
  pyannote-face landmarks [--verbose] <video> <model> <tracking> <output>
  pyannote-face features [--verbose] <video> <model> <landmark> <output>
  pyannote-face demo [--from=<sec>] [--until=<sec>] [--shift=<sec>] [--label=<path>] <video> <tracking> <output>
  pyannote-face (-h | --help)
  pyannote-face --version

Options:
  --smallest=<size>         (Approximate) size of smallest face [default: 36].
  --min-overlap=<ratio>     Associates face with tracker if overlap is greater
                            than <ratio> [default: 0.5].
  --min-confidence=<float>  Reset trackers with confidence lower than <float>
                            [default: 10.].
  --max-gap=<float>         Bridge gaps with duration shorter than <float>
                            [default: 1.].
  --from=<sec>              Encode demo from <sec> seconds [default: 0].
  --until=<sec>             Encode demo until <sec> seconds.
  --shift=<sec>             Shift tracks by <sec> seconds [default: 0].
  --label=<path>            Track labels.
  -h --help                 Show this screen.
  --version                 Show version.
  --verbose                 Show progress.
"""

from docopt import docopt

from pyannote.core import Annotation
from pyannote.core.json import load

from pyannote.video import __version__
from pyannote.video import Video
from pyannote.video import Face
from pyannote.video import FaceTracking

from pandas import read_table

from six.moves import zip
import numpy as np
import cv2

import dlib


SMALLEST_DEFAULT = 36
MIN_OVERLAP_RATIO = 0.5
MIN_CONFIDENCE = 10.
MAX_GAP = 1.

FACE_TEMPLATE = ('{t:.3f} {identifier:d} '
                 '{left:d} {top:d} {right:d} {bottom:d} '
                 '{status:s}\n')


def getFaceGenerator(tracking, double=True):
    """Parse precomputed face file and generate timestamped faces"""

    # load tracking file and sort it by timestamp
    names = ['t', 'track', 'left', 'top', 'right', 'bottom', 'status']
    dtype = {'left': np.int32, 'top': np.int32,
             'right': np.int32, 'bottom': np.int32}
    tracking = read_table(tracking, delim_whitespace=True, header=None,
                          names=names, dtype=dtype)
    tracking = tracking.sort_values('t')

    # t is the time sent by the frame generator
    t = yield

    rectangle = dlib.drectangle if double else dlib.rectangle

    faces = []
    currentT = None

    for _, (T, identifier, left, top, right, bottom, status) in tracking.iterrows():

        face = rectangle(left, top, right, bottom)

        # load all faces from current frame and only those faces
        if T == currentT or currentT is None:
            faces.append((identifier, face, status))
            currentT = T
            continue

        # once all faces at current time are loaded
        # wait until t reaches current time
        # then returns all faces at once

        while True:

            # wait...
            if currentT > t:
                t = yield t, []
                continue

            # return all faces at once
            t = yield currentT, faces

            # reset current time and corresponding faces
            faces = [(identifier, face, status)]
            currentT = T
            break

    while True:
        t = yield t, []


def pairwise(iterable):
    "s -> (s0,s1), (s2,s3), (s4, s5), ..."
    a = iter(iterable)
    return zip(a, a)


def getShapeGenerator(shape):
    """Parse precomputed shape file and generate timestamped shapes"""

    # load landmarks file
    shape = read_table(shape, delim_whitespace=True, header=None)

    # deduce number of landmarks from file dimension
    _, d = shape.shape
    n_points = (d - 2) / 2

    # t is the time sent by the frame generator
    t = yield

    shapes = []
    currentT = None

    for _, row in shape.iterrows():

        T = float(row[0])
        identifier = int(row[1])
        landmarks = np.float32(list(pairwise(
            [int(coordinate) for coordinate in row[2:]])))

        # load all shapes from current frame
        # and only those shapes
        if T == currentT or currentT is None:
            shapes.append((identifier, landmarks))
            currentT = T
            continue

        # once all shapes at current time are loaded
        # wait until t reaches current time
        # then returns all shapes at once

        while True:

            # wait...
            if currentT > t:
                t = yield t, []
                continue

            # return all shapes at once
            t = yield currentT, shapes

            # reset current time and corresponding shapes
            shapes = [(identifier, landmarks)]
            currentT = T
            break

    while True:
        t = yield t, []


def track(video, shot, output, smallest=36,
          min_overlap_ratio=MIN_OVERLAP_RATIO, min_confidence=MIN_CONFIDENCE,
          max_gap=MAX_GAP):
    """Tracking by detection"""

    tracking = FaceTracking(smallest=smallest,
                            min_overlap_ratio=min_overlap_ratio,
                            min_confidence=min_confidence,
                            max_gap=max_gap)

    shot = load(shot)

    if isinstance(shot, Annotation):
        shot = shot.get_timeline()

    with open(output, 'w') as foutput:

        for identifier, track in enumerate(tracking(video, shot)):

            for t, (left, top, right, bottom), status in track:

                foutput.write(FACE_TEMPLATE.format(
                    t=t, identifier=identifier, status=status,
                    left=left, right=right, top=top, bottom=bottom))

            foutput.flush()

def landmark(video, model, tracking, output):
    """Facial features detection"""

    # face generator
    faceGenerator = getFaceGenerator(tracking, double=False)
    faceGenerator.send(None)

    face = Face(landmarks=model)

    with open(output, 'w') as foutput:

        for timestamp, rgb in video:

            # get all detected faces at this time
            T, faces = faceGenerator.send(timestamp)
            # not that T might be differ slightly from t
            # due to different steps in frame iteration

            for identifier, boundingBox, _ in faces:

                landmarks = face._get_landmarks(rgb, boundingBox)

                foutput.write('{t:.3f} {identifier:d}'.format(
                    t=T, identifier=identifier))
                for x, y in landmarks:
                    foutput.write(' {x:d} {y:d}'.format(x=int(x), y=int(y)))
                foutput.write('\n')

            foutput.flush()

def features(video, model, shape, output):
    """Openface FaceNet feature extraction"""

    face = Face(size=96, openface=model)

    # shape generator
    shapeGenerator = getShapeGenerator(shape)
    shapeGenerator.send(None)

    with open(output, 'w') as foutput:

        for timestamp, rgb in video:

            T, shapes = shapeGenerator.send(timestamp)

            for identifier, landmarks in shapes:
                normalized_rgb = face._get_normalized(rgb, landmarks)
                normalized_bgr = cv2.cvtColor(normalized_rgb,
                                              cv2.COLOR_BGR2RGB)
                openface = face._get_openface(normalized_bgr)

                foutput.write('{t:.3f} {identifier:d}'.format(
                    t=T, identifier=identifier))
                for x in openface:
                    foutput.write(' {x:.5f}'.format(x=x))
                foutput.write('\n')

            foutput.flush()

def get_fl(tracking, shift=0., labels=None):

    COLORS = [
        (240, 163, 255), (  0, 117, 220), (153,  63,   0), ( 76,   0,  92),
        ( 25,  25,  25), (  0,  92,  49), ( 43, 206,  72), (255, 204, 153),
        (128, 128, 128), (148, 255, 181), (143, 124,   0), (157, 204,   0),
        (194,   0, 136), (  0,  51, 128), (255, 164,   5), (255, 168, 187),
        ( 66, 102,   0), (255,   0,  16), ( 94, 241, 242), (  0, 153, 143),
        (224, 255, 102), (116,  10, 255), (153,   0,   0), (255, 255, 128),
        (255, 255,   0), (255,  80,   5)
    ]

    faceGenerator = getFaceGenerator(tracking, double=True)
    faceGenerator.send(None)

    if labels is None:
        labels = dict()

    def overlay(get_frame, timestamp):
        frame = get_frame(timestamp)
        height, width, _ = frame.shape
        _, faces = faceGenerator.send(timestamp - shift)

        cv2.putText(frame, '{t:.3f}'.format(t=timestamp), (10, height-10),
                    cv2.FONT_HERSHEY_DUPLEX, 0.5, (255, 0, 0), 1, 8, False)
        for identifier, face, _ in faces:
            color = COLORS[identifier % len(COLORS)]

            # Draw face bounding box
            pt1 = (int(face.left()), int(face.top()))
            pt2 = (int(face.right()), int(face.bottom()))
            cv2.rectangle(frame, pt1, pt2, color, 2)

            # Print tracker identifier
            cv2.putText(frame, '#{identifier:d}'.format(identifier=identifier),
                        (pt1[0], pt2[1] + 15), cv2.FONT_HERSHEY_SIMPLEX,
                        0.5, (255, 0, 0), 1, 8, False)

            # Print track label
            label = labels.get(identifier, '')
            cv2.putText(frame,
                        '{label:s}'.format(label=label),
                        (pt1[0], pt1[1] - 7), cv2.FONT_HERSHEY_SIMPLEX,
                        0.5, (255, 0, 0), 1, 8, False)
        return frame

    return overlay


def demo(filename, tracking, output, t_start=0., t_end=None, shift=0.,
         labels=None):

    import os
    os.environ['IMAGEIO_FFMPEG_EXE'] = 'ffmpeg'
    from moviepy.video.io.VideoFileClip import VideoFileClip

    if labels is not None:
        with open(labels, 'r') as f:
            labels = {}
            for line in f:
                identifier, label = line.strip().split()
                identifier = int(identifier)
                labels[identifier] = label

    original_clip = VideoFileClip(filename)
    modified_clip = original_clip.fl(get_fl(tracking,
                                            shift=shift,
                                            labels=labels))
    cropped_clip = modified_clip.subclip(t_start=t_start, t_end=t_end)
    cropped_clip.write_videofile(output)


if __name__ == '__main__':

    # parse command line arguments
    version = 'pyannote-face {version}'.format(version=__version__)
    arguments = docopt(__doc__, version=version)

    # initialize video
    filename = arguments['<video>']

    verbose = arguments['--verbose']

    video = Video(filename, verbose=verbose)

    # face tracking
    if arguments['track']:

        shot = arguments['<shot.json>']
        output = arguments['<output>']
        min_overlap_ratio = float(arguments['--min-overlap'])
        min_confidence = float(arguments['--min-confidence'])
        max_gap = float(arguments['--max-gap'])
        track(video, shot, output,
              min_overlap_ratio=min_overlap_ratio,
              min_confidence=min_confidence,
              max_gap=max_gap)

    # facial features detection
    if arguments['landmarks']:

        tracking = arguments['<tracking>']
        model = arguments['<model>']
        output = arguments['<output>']
        landmark(video, model, tracking, output)

    # openface features extraction
    if arguments['features']:

        model = arguments['<model>']
        shape = arguments['<landmark>']
        output = arguments['<output>']
        features(video, model, shape, output)

    if arguments['demo']:

        tracking = arguments['<tracking>']
        output = arguments['<output>']

        t_start = float(arguments['--from'])
        t_end = arguments['--until']
        t_end = float(t_end) if t_end else None

        shift = float(arguments['--shift'])
        labels = arguments['--label']
        if not labels:
            labels = None

        demo(filename, tracking, output,
             t_start=t_start, t_end=t_end,
             shift=shift, labels=labels)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2015-2016 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr


from __future__ import print_function
from pyannote.core import Annotation
from pyannote.core import Segment
from pyannote.core import Unknown
from pyannote.core.time import _t_iter as getLabelGenerator
import sklearn.cluster
import numpy as np


class Cluster(object):

    def __init__(self, min_duration=1.0, min_size=60):
        super(Cluster, self).__init__()
        self.min_duration = min_duration
        self.min_size = min_size

    def _load(self, features, tracking):

        # track duration
        groups = features.groupby('track')

        # track timespan
        timespans = groups.apply(lambda x: Segment(np.min(x.time), np.max(x.time)))

        # track duration
        duration = groups.apply(lambda faces: np.max(faces.time) - np.min(faces.time))

        # track size
        size = tracking.groupby('track').apply(lambda faces: np.mean([face.right-face.left for _, face in faces.iterrows()]))

        # track descriptor (average Openface features)
        descriptor = groups.mean().drop('time', axis=1)

        # get rid of short and small tracklets
        keep = (duration > self.min_duration) & (size > self.min_size)
        kept_descriptor = descriptor[keep]
        kept_timespans = timespans[keep]
        X = np.array(kept_descriptor)

        return X, kept_timespans

    def _cluster_to_annotation(self, clusters, kept_timespans):

        labelGenerator = getLabelGenerator()
        annotation = Annotation()

        for cluster in np.unique(clusters):

            label = Unknown() if cluster == -1 else next(labelGenerator)

            for track, segment in kept_timespans[clusters == cluster].iteritems():
                annotation[segment, track] = label

        return annotation

    def __call__(self, features, tracking, n_clusters=50):

        X, kept_timespans = self._load(features, tracking)

        annotation = Annotation()
        if len(X) == 0:
            return annotation

        # # precompute pairwise euclidean distance
        # S = np.sqrt(((X[np.newaxis, :, :] - X[:, np.newaxis, :]) ** 2).sum(2))

        # ward = sklearn.cluster.AgglomerativeClustering(
        #     n_clusters=n_clusters, affinity='euclidean', linkage='ward')
        # clusters = ward.fit_predict(X)

        affinityPropagation = sklearn.cluster.AffinityPropagation(
            damping=0.5, max_iter=200, convergence_iter=15, copy=True,
            preference=None, affinity='euclidean', verbose=True)
        clusters = affinityPropagation.fit_predict(X)

        # birch = sklearn.cluster.Birch(threshold=0.2, n_clusters=n_clusters)
        # clusters = birch.fit_predict(X)

        return self._cluster_to_annotation(clusters, kept_timespans)

    def visualize(self, video, annotation, tracking):

        import scipy.misc

        faces = tracking.groupby('track').first()

        mugshots = {}

        for i, (label, _) in enumerate(annotation.chart()):

            print(i)

            mugshots[label] = []

            for _, track in annotation.subset(set([label])).itertracks():

                face = faces.xs(track)
                t = face.time

                left, right = max(0, int(face.left)), int(face.right)
                top, bottom = max(0, int(face.top)), int(face.bottom)

                print(t, left, right, top, bottom)

                frame = video(t)
                cropped = frame[top:bottom, left:right, :]
                mugshot = scipy.misc.imresize(cropped, (60, 60), interp='bilinear')

                mugshots[label].append(mugshot)

            mugshots[label] = np.hstack(mugshots[label])

        return mugshots



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2015-2016 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

import cv2
import itertools
import numpy as np
from ..video import Video
import scipy.signal
from pyannote.core import Segment

OPENCV = int(cv2.__version__.split('.')[0])


class Shot(object):
    """Shot boundary detection based on displaced frame difference

    Parameters
    ----------
    video : Video
    height : int, optional
        Resize video to this height, in pixels. Defaults to 50.
    context : float, optional
        Median filtering context (in seconds). Defaults to 2.
    threshold : float, optional
        Defaults to 1.
    """

    def __init__(self, video, height=50, context=2.0, threshold=1.0):
        super(Shot, self).__init__()
        self.video = video
        self.height = height
        self.threshold = threshold
        self.context = context

        # estimate new size from video size and target height
        w, h = self.video._size
        self._resize = (self.height, w * self.height / h)

        # estimate kernel size from context and video step
        kernel_size = self.context / self.video.step
        # kernel size must be an odd number greater than 3
        self._kernel_size = max(3, int(np.ceil(kernel_size) // 2 * 2 + 1))

        self._reconstruct = None

    def _convert(self, rgb):
        gray = cv2.cvtColor(rgb, cv2.COLOR_RGB2GRAY)
        return cv2.resize(gray, self._resize)

    def dfd(self, previous, current, flow=None):
        """Displaced frame difference"""

        if OPENCV == 2:
            flow = cv2.calcOpticalFlowFarneback(
                previous, current, 0.5, 3, 15, 3, 5, 1.1, 0, flow)
        elif OPENCV == 3:
            flow = cv2.calcOpticalFlowFarneback(
                previous, current, flow, 0.5, 3, 15, 3, 5, 1.1, 0)

        height, width = previous.shape

        # allocate "reconstruct" only once
        if self._reconstruct is None:
            self._reconstruct = np.empty(previous.shape)

        for x, y in itertools.product(range(width), range(height)):
            dy, dx = flow[y, x]
            rx = max(0, min(x + dx, width - 1))
            ry = max(0, min(y + dy, height - 1))
            self._reconstruct[y, x] = current[ry, rx]

        return np.mean(np.abs(previous - self._reconstruct))

    def iter_dfd(self):
        """Pairwise displaced frame difference"""

        previous = None

        # iterate frames one by one
        for t, rgb in self.video:

            current = self._convert(rgb)

            if previous is None:
                previous = current
                continue

            yield t, self.dfd(previous, current, flow=None)

            previous = current

    def __iter__(self):

        # TODO: running median
        t, y = zip(*self.iter_dfd())

        filtered = scipy.signal.medfilt(y, kernel_size=self._kernel_size)

        # normalized displaced frame difference
        normalized = (y - filtered) / filtered

        # apply threshold on normalized displaced frame difference
        # in case multiple consecutive value are higher than the threshold,
        # only keep the first one as a shot boundary.
        previous = self.video.start
        _i = 0
        for i in np.where(normalized > self.threshold)[0]:

            if i == _i + 1:
                _i = i
                continue

            yield Segment(previous, t[i])

            previous = t[i]
            _i = i

        yield Segment(previous, self.video.end)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2016 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

"""Face tracking"""

from .face import Face
from .face import SMALLEST_DEFAULT
from ..tracking import TrackingByDetection


def get_face_detect(face):
    """Create function for face detection"""
    def face_detect(frame):
        """Detect face in frame"""
        for f in face.iterfaces(frame):
            yield (f.left(), f.top(), f.right(), f.bottom())
    return face_detect


class FaceTracking(TrackingByDetection):
    """Face tracking

    Parameters
    ----------
    smallest : int, optional
        Approximate size in pixel of the smallest face that should be
        detected. Defaults to 36.
    min_confidence : float, optional
        Kill trackers whose confidence goes below this value. Defaults to 10.
    min_overlap_ratio : float, optional
        Do not associate trackers and detections if their overlap ratio goes
        below this value. Defaults to 0.3.
    max_gap : float, optional
        Bridge gaps with duration shorter than this value.
    """
    def __init__(self, smallest=SMALLEST_DEFAULT,
                 min_confidence=10., min_overlap_ratio=0.3, max_gap=0.):

        face = Face(smallest=smallest)
        detect_func = get_face_detect(face)

        super(FaceTracking, self).__init__(
            detect_func=detect_func,
            min_confidence=min_confidence,
            min_overlap_ratio=min_overlap_ratio,
            max_gap=max_gap)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2015-2016 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr


"""Video structure

The standard pipeline for is the following:

    shot boundary detection ==> shot threading ==> segmentation into scenes

Usage:
  pyannote-structure.py shot [options] <video> <output.json>
  pyannote-structure.py thread [options] <video> <shot.json> <output.json>
  pyannote-structure.py scene [options] <video> <thread.json> <output.json>
  pyannote-structure.py (-h | --help)
  pyannote-structure.py --version

Options:
  --height=<n_pixels>    Resize video frame to height <n_pixels> [default: 50].
  --window=<n_seconds>   Apply median filtering on <n_seconds> window [default: 2.0].
  --threshold=<value>    Set threshold to <value> [default: 1.0].
  --min-match=<n_match>  Set minimum number of matches to <n_match> [default: 20].
  --lookahead=<n_shots>  Look at up to <n_shots> following shots [default: 24].
  -h --help              Show this screen.
  --version              Show version.
  --verbose              Show progress.
"""

from docopt import docopt

from pyannote.core import Timeline
from pyannote.core.json import dump, load

from pyannote.video import __version__
from pyannote.video import Video
from pyannote.video import Shot, Thread


def do_shot(video, output, height=50, window=2.0, threshold=1.0):

    shots = Shot(video, height=height, context=window, threshold=threshold)
    shots = Timeline(shots)
    dump(shots, output)

def do_thread(video, shots, output, min_match=20, lookahead=24, verbose=False):

    shots = load(shots)
    threads = Thread(video, shot=shots, lookahead=lookahead,
                     min_match=min_match, verbose=verbose)
    threads = threads()
    dump(threads, output)

def do_scene(video, threads, output, verbose=False):

    threads = load(threads)
    raise NotImplementedError('Not yet available')
    # scenes = Scene(video, thread=threads, verbose=verbose)
    # dump(scenes, output)


if __name__ == '__main__':

    # parse command line arguments
    version = 'pyannote-structure {version}'.format(version=__version__)
    arguments = docopt(__doc__, version=version)

    # common 'verbosity' option
    verbose = arguments['--verbose']

    # common 'output' argument
    output = arguments['<output.json>']

    # initialize video
    filename = arguments['<video>']
    video = Video(filename, verbose=verbose)

    if arguments['shot']:
        height = int(arguments['--height'])
        window = float(arguments['--window'])
        threshold = float(arguments['--threshold'])
        do_shot(video, output,
                height=height, window=window, threshold=threshold)

    if arguments['thread']:
        shots = arguments['<shot.json>']
        min_match = int(arguments['--min-match'])
        lookahead = int(arguments['--lookahead'])
        do_thread(video, shots, output,
                  min_match=min_match, lookahead=lookahead, verbose=verbose)

    if arguments['scene']:
        threads = arguments['<thread.json>']
        do_scene(video, threads, output, verbose=verbose)



# Most code in this file was shamelessly borrowed from MoviePy
# http://zulko.github.io/moviepy/

# The MIT License (MIT)
#
# Copyright (c) 2015 Zulko
# Copyright (c) 2015-2016 CNRS
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
This module implements all the functions to read a video or a picture
using ffmpeg. It is quite ugly, as there are many pitfalls to avoid
"""

from __future__ import division

import subprocess as sp
import os
import re
import warnings
import logging
import numpy as np
from collections import deque
from tqdm import tqdm

logging.captureWarnings(True)

try:
    from subprocess import DEVNULL  # py3k
except ImportError:
    DEVNULL = open(os.devnull, 'wb')


def _is_string(obj):
    """ Returns true if s is string or string-like object,
    compatible with Python 2 and Python 3."""
    try:
        return isinstance(obj, basestring)
    except NameError:
        return isinstance(obj, str)


def _cvsecs(time):
    """ Will convert any time into seconds.
    Here are the accepted formats:
    >>> _cvsecs(15.4) -> 15.4 # seconds
    >>> _cvsecs( (1,21.5) ) -> 81.5 # (min,sec)
    >>> _cvsecs( (1,1,2) ) -> 3662 # (hr, min, sec)
    >>> _cvsecs('01:01:33.5') -> 3693.5  #(hr,min,sec)
    >>> _cvsecs('01:01:33.045') -> 3693.045
    >>> _cvsecs('01:01:33,5') #coma works too
    """

    if _is_string(time):
        if (',' not in time) and ('.' not in time):
            time = time + '.0'
        expr = r"(\d+):(\d+):(\d+)[,|.](\d+)"
        finds = re.findall(expr, time)[0]
        nums = list(map(float, finds))
        return (3600*int(finds[0]) +
                60*int(finds[1]) +
                int(finds[2]) +
                nums[3]/(10**len(finds[3])))

    elif isinstance(time, tuple):
        if len(time) == 3:
            hr, mn, sec = time
        elif len(time) == 2:
            hr, mn, sec = 0, time[0], time[1]
        return 3600*hr + 60*mn + sec

    else:
        return time


class Video:

    def __init__(self, filename, start=None, end=None, step=None,
                 ffmpeg='ffmpeg', verbose=False):
        """
        Parameters
        ----------
        start : float, optional
            Begin iterating frames at time `start` (in seconds).
            Defaults to 0.
        end : float, optional
            Stop iterating frames at time `end` (in seconds).
            Defaults to video duration.
        step : float, optional
            Iterate frames every `step` seconds.
            Defaults to iterating every frame.
        verbose : bool, optional
            Show a progress bar while iterating the video. Defaults to False.
        ffmpeg : str, optional
            Path to ffmpeg command line tool. Defaults to "ffmpeg".
        """

        self.filename = filename
        self.ffmpeg = ffmpeg
        self.verbose = verbose

        infos = self._parse_infos(print_infos=False, check_duration=True)
        self._fps = infos['video_fps']
        self._size = infos['video_size']
        self._duration = infos['video_duration']
        # self.ffmpeg_duration = infos['duration']
        self._nframes = infos['video_nframes']

        self.start = 0. if start is None else start
        self.end = self._duration if end is None else end
        self.step = 1./self._fps if step is None else step

        # TODO warning if step != N x 1/fps (where N is an integer)
        # warnings.warn(message, UserWarning)

        # self.infos = infos

        self._pix_fmt = 'rgb24'
        self._depth = 3

        w, h = self._size
        bufsize = self._depth * w * h + 100

        self._bufsize = bufsize
        self._initialize()

        self._pos = 1
        self._lastread = self._read_frame()

    def _parse_infos(self, print_infos=False, check_duration=True):
        """Get file infos using ffmpeg.

        Returns a dictionnary with the fields:
        "video_found", "video_fps", "duration", "video_nframes",
        "video_duration", "audio_found", "audio_fps"

        "video_duration" is slightly smaller than "duration" to avoid
        fetching the uncomplete frames at the end, which raises an error.

        """

        # open the file in a pipe, provoke an error, read output
        is_GIF = self.filename.endswith('.gif')
        cmd = [self.ffmpeg, "-i", self.filename]
        if is_GIF:
            cmd += ["-f", "null", "/dev/null"]

        popen_params = {"bufsize": 10**5,
                        "stdout": sp.PIPE,
                        "stderr": sp.PIPE,
                        "stdin": DEVNULL}

        if os.name == "nt":
            popen_params["creationflags"] = 0x08000000

        proc = sp.Popen(cmd, **popen_params)

        proc.stdout.readline()
        proc.terminate()
        infos = proc.stderr.read().decode('utf8')
        del proc

        if print_infos:
            # print the whole info text returned by FFMPEG
            print(infos)

        lines = infos.splitlines()
        if "No such file or directory" in lines[-1]:
            raise IOError(
                ("MoviePy error: the file %s could not be found !\n"
                 "Please check that you entered the correct path.") % self.filename)

        result = dict()

        # get duration (in seconds)
        result['duration'] = None

        if check_duration:
            try:
                keyword = ('frame=' if is_GIF else 'Duration: ')
                line = [l for l in lines if keyword in l][0]
                match = re.findall(
                    "([0-9][0-9]:[0-9][0-9]:[0-9][0-9].[0-9][0-9])", line)[0]
                result['duration'] = _cvsecs(match)
            except:
                raise IOError(
                    ("MoviePy error: failed to read the duration of file %s.\n"
                     "Here are the file infos returned by ffmpeg:\n\n%s") % (
                                  self.filename, infos))

        # get the output line that speaks about video
        lines_video = [l for l in lines
                       if ' Video: ' in l and re.search('\d+x\d+', l)]

        result['video_found'] = lines_video != []

        if result['video_found']:

            line = lines_video[0]

            try:

                # get the size, of the form 352x288 (w x h)
                m = re.search(" [0-9]*x[0-9]*(,| )", line)
                s = [int(p) for p in line[m.start():m.end()-1].split('x')]
                result['video_size'] = s

            except:
                message = (
                    "MoviePy error: failed to read video dimensions in file "
                    "%s.\nHere are the file infos returned by ffmpeg:\n\n%s"
                )
                raise IOError(message % (self.filename, infos))

            try:

                # get the display aspect ratio, of the form 178:163
                m = re.search("DAR [0-9]*:[0-9]*", line)
                s = [int(p) for p in line[m.start()+4:m.end()].split(':')]
                result['video_dar'] = s

            except:
                result['video_dar'] = None

            # get the frame rate. Sometimes it's 'tbr', sometimes 'fps',
            # sometimes tbc, and sometimes tbc/2...
            # Current policy: Trust tbr first, then fps. If result is near
            # from x*1000/1001 where x is 23,24,25,50, replace by x*1000/1001
            # (very common case for the fps).

            try:
                match = re.search("( [0-9]*.| )[0-9]* tbr", line)
                tbr = float(line[match.start():match.end()].split(' ')[1])
                result['video_fps'] = tbr

            except:
                match = re.search("( [0-9]*.| )[0-9]* fps", line)
                result['video_fps'] = float(line[match.start():match.end()].split(' ')[1])

            # It is known that a fps of 24 is often written as 24000/1001
            # but then ffmpeg nicely rounds it to 23.98, which we hate.
            coef = 1000.0/1001.0
            fps = result['video_fps']
            for x in [23, 24, 25, 30, 50]:
                if (fps!=x) and abs(fps - x*coef) < .01:
                    result['video_fps'] = x*coef

            if check_duration:
                result['video_nframes'] = int(result['duration']*result['video_fps'])+1
                result['video_duration'] = result['duration']
            else:
                result['video_nframes'] = 1
                result['video_duration'] = None

            # We could have also recomputed the duration from the number
            # of frames, as follows:
            # >>> result['video_duration'] = result['video_nframes'] / result['video_fps']

        lines_audio = [l for l in lines if ' Audio: ' in l]

        result['audio_found'] = lines_audio != []

        if result['audio_found']:
            line = lines_audio[0]
            try:
                match = re.search(" [0-9]* Hz", line)
                result['audio_fps'] = int(line[match.start()+1:match.end()])
            except:
                result['audio_fps'] = 'unknown'

        return result

    def _initialize(self, t=0):
        """Opens the file, creates the pipe. """

        self._close()  # if any

        if t != 0:
            offset = min(1, t)
            i_arg = ['-ss', "%.06f" % (t - offset),
                     '-i', self.filename,
                     '-ss', "%.06f" % offset]
        else:
            i_arg = ['-i', self.filename]

        cmd = (
            [self.ffmpeg] + i_arg +
            ['-loglevel', 'error', '-f', 'image2pipe',
             '-pix_fmt', self._pix_fmt, '-vcodec', 'rawvideo', '-'])

        popen_params = {"bufsize": self._bufsize,
                        "stdout": sp.PIPE,
                        "stderr": sp.PIPE,
                        "stdin": DEVNULL}

        if os.name == "nt":
            popen_params["creationflags"] = 0x08000000

        self._proc = sp.Popen(cmd, **popen_params)

    def _skip_frames(self, n=1):
        """Reads and throws away n frames """
        w, h = self._size
        for _ in range(n):
            self._proc.stdout.read(self._depth*w*h)
            # self._proc.stdout.flush()
        self._pos += n

    def _read_frame(self):
        w, h = self._size
        nbytes = self._depth*w*h

        s = self._proc.stdout.read(nbytes)

        if len(s) != nbytes:

            warnings.warn(
                "Warning: in file %s, " % (self.filename) +
                "%d bytes wanted but %d bytes read," % (nbytes, len(s)) +
                "at frame %d/%d, at time %.02f/%.02f sec. " % (
                    self._pos, self._nframes, 1.0 * self._pos / self._fps,
                    self._duration) +
                "Using the last valid frame instead.",
                UserWarning)

            if not hasattr(self, '_lastread'):
                message = (
                    "MoviePy error: failed to read the first frame of "
                    "video file %s. That might mean that the file is "
                    "corrupted. That may also mean that you are using "
                    "a deprecated version of FFMPEG. On Ubuntu/Debian "
                    "for instance the version in the repos is deprecated. "
                    "Please update to a recent version from the website."
                )
                raise IOError(message % self.filename)

            result = self._lastread

        else:

            result = np.fromstring(s, dtype='uint8')
            # reshape((h, w, len(s)//(w*h)))
            result.shape = (h, w, len(s)//(w*h))
            self._lastread = result

        return result

    def __iter__(self):
        return self.iterframes(with_time=True)

    def iterframes(self, with_time=False, with_context=False, context=1):
        """Iterate over video frames

        Frames are generated as H x W x 3 numpy array in RGB order (not BGR).
        (FYI, OpenCV standard format is BGR, not RGB).

        Parameters
        ----------
        with_time : boolean
            When True, yields (time, frame).
        with_context : {False, 'left', 'right', 'center'}
            Defaults to False.
        context : int
            Number of contextual frames. Defaults to 1.
        """

        # initialize buffer of contextual frames
        if with_context:
            frames = deque([], context)
            timestamps = deque([], context)

        generator = np.arange(self.start, self.end, self.step)
        if self.verbose:
            generator = tqdm(iterable=generator,
                             total=(self.end - self.start) / self.step,
                             leave=True, mininterval=1.,
                             unit='frames', unit_scale=True)

        for t in generator:

            rgb = self._get_frame(t)

            # fill buffer of contextual frames
            if with_context:
                frames.append(rgb)
                timestamps.append(t)
                if len(frames) < context:
                    continue

            f_ = frames if with_context else rgb

            if with_time:

                if with_context is 'right':
                    t_ = timestamps[0]
                elif with_context is 'center':
                    t_ = timestamps[context / 2]
                else:
                    t_ = t

                yield t_, f_

            else:
                yield f_

    def __call__(self, t):
        return self._get_frame(t)

    def _get_frame(self, t):
        """ Read a file video frame at time t.

        Note for coders: getting an arbitrary frame in the video with
        ffmpeg can be painfully slow if some decoding has to be done.
        This function tries to avoid fectching arbitrary frames
        whenever possible, by moving between adjacent frames.
        """

        # these definitely need to be rechecked sometime. Seems to work.

        # I use that horrible '+0.00001' hack because sometimes due to
        # numerical imprecisions a 3.0 can become a 2.99999999... which makes
        # the int() go to the previous integer. This makes the fetching more
        # robust in the case where you get the nth frame by writing
        # _get_frame(n/fps).

        pos = int(self._fps * t + 0.00001)+1

        if pos == self._pos:
            return self._lastread
        else:
            if(pos < self._pos) or (pos > self._pos+100):
                self._initialize(t)
                self._pos = pos
            else:
                self._skip_frames(pos-self._pos-1)
            result = self._read_frame()
            self._pos = pos
            return result

    def _close(self):
        if hasattr(self, '_proc'):
            self._proc.terminate()
            self._proc.stdout.close()
            self._proc.stderr.close()
            del self._proc

    def __del__(self):
        self._close()
        if hasattr(self, '_lastread'):
            del self._lastread



# Copyright 2015 Carnegie Mellon University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
import binascii
from subprocess import Popen, PIPE
import os
import os.path
import sys

import numpy as np
import cv2


class TorchWrap(object):
    # Warning: This is very unstable!
    # Please join us in improving it at:
    #   https://github.com/cmusatyalab/openface/issues/1
    #   https://github.com/cmusatyalab/openface/issues/4

    def __init__(self, torch='/root/torch/install/bin/th',
                 model='nn4.v2.t7', size=96, cuda=False):

        super(TorchWrap, self).__init__()

        luaFile = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            'openface_server.lua')

        self.cmd = [torch, luaFile, '-model', model, '-imgDim', str(size)]

        if cuda:
            self.cmd.append('-cuda')

        self.p = Popen(self.cmd, stdin=PIPE, stdout=PIPE,
                       stderr=PIPE, bufsize=0)

        def exitHandler():
            if self.p.poll() is None:
                self.p.kill()

        atexit.register(exitHandler)

    def forwardPath(self, imgPath):
        rc = self.p.poll()
        if rc is not None and rc != 0:
            raise Exception("""


OpenFace: `openface_server.lua` subprocess has died.
Is the Torch command `th` on your PATH? Check with `which th`.

Diagnostic information:

cmd: {}

============

stdout: {}

============

stderr: {}
""".format(self.cmd, self.p.stdout.read(), self.p.stderr.read()))

        self.p.stdin.write(imgPath + "\n")
        output = self.p.stdout.readline()
        try:
            return [float(x) for x in output.strip().split(',')]
        except Exception as e:
            self.p.kill()
            stdout, stderr = self.p.communicate()
            print("""


Error getting result from Torch subprocess.

Line read: {}

Exception:

{}

============

stdout: {}

============

stderr: {}
""".format(output, str(e), stdout, stderr))
            sys.exit(-1)

    def forwardImage(self, bgr):
        t = '/tmp/openface-torchwrap-{}.png'.format(
            binascii.b2a_hex(os.urandom(8)))
        cv2.imwrite(t, bgr)
        rep = np.array(self.forwardPath(t))
        os.remove(t)
        return rep



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2015-2016 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

"""Face processing"""


import numpy as np
import dlib
import cv2
from .openface import TorchWrap

SMALLEST_DEFAULT = 36

TEMPLATE = np.float32([
    (0.0792396913815, 0.339223741112), (0.0829219487236, 0.456955367943),
    (0.0967927109165, 0.575648016728), (0.122141515615, 0.691921601066),
    (0.168687863544, 0.800341263616), (0.239789390707, 0.895732504778),
    (0.325662452515, 0.977068762493), (0.422318282013, 1.04329000149),
    (0.531777802068, 1.06080371126), (0.641296298053, 1.03981924107),
    (0.738105872266, 0.972268833998), (0.824444363295, 0.889624082279),
    (0.894792677532, 0.792494155836), (0.939395486253, 0.681546643421),
    (0.96111933829, 0.562238253072), (0.970579841181, 0.441758925744),
    (0.971193274221, 0.322118743967), (0.163846223133, 0.249151738053),
    (0.21780354657, 0.204255863861), (0.291299351124, 0.192367318323),
    (0.367460241458, 0.203582210627), (0.4392945113, 0.233135599851),
    (0.586445962425, 0.228141644834), (0.660152671635, 0.195923841854),
    (0.737466449096, 0.182360984545), (0.813236546239, 0.192828009114),
    (0.8707571886, 0.235293377042), (0.51534533827, 0.31863546193),
    (0.516221448289, 0.396200446263), (0.517118861835, 0.473797687758),
    (0.51816430343, 0.553157797772), (0.433701156035, 0.604054457668),
    (0.475501237769, 0.62076344024), (0.520712933176, 0.634268222208),
    (0.565874114041, 0.618796581487), (0.607054002672, 0.60157671656),
    (0.252418718401, 0.331052263829), (0.298663015648, 0.302646354002),
    (0.355749724218, 0.303020650651), (0.403718978315, 0.33867711083),
    (0.352507175597, 0.349987615384), (0.296791759886, 0.350478978225),
    (0.631326076346, 0.334136672344), (0.679073381078, 0.29645404267),
    (0.73597236153, 0.294721285802), (0.782865376271, 0.321305281656),
    (0.740312274764, 0.341849376713), (0.68499850091, 0.343734332172),
    (0.353167761422, 0.746189164237), (0.414587777921, 0.719053835073),
    (0.477677654595, 0.706835892494), (0.522732900812, 0.717092275768),
    (0.569832064287, 0.705414478982), (0.635195811927, 0.71565572516),
    (0.69951672331, 0.739419187253), (0.639447159575, 0.805236879972),
    (0.576410514055, 0.835436670169), (0.525398405766, 0.841706377792),
    (0.47641545769, 0.837505914975), (0.41379548902, 0.810045601727),
    (0.380084785646, 0.749979603086), (0.477955996282, 0.74513234612),
    (0.523389793327, 0.748924302636), (0.571057789237, 0.74332894691),
    (0.672409137852, 0.744177032192), (0.572539621444, 0.776609286626),
    (0.5240106503, 0.783370783245), (0.477561227414, 0.778476346951)])

TPL_MIN, TPL_MAX = np.min(TEMPLATE, axis=0), np.max(TEMPLATE, axis=0)
MINMAX_TEMPLATE = (TEMPLATE - TPL_MIN) / (TPL_MAX - TPL_MIN)

OUTER_EYES_AND_NOSE = np.array([36, 45, 33])

class Face(object):
    """Face processing"""
    def __init__(self, landmarks=None, smallest=SMALLEST_DEFAULT,
                 openface=None, size=96):
        """Face detection

        Parameters
        ----------
        landmarks : str
            Path to dlib's 68 facial landmarks predictor model.
        smallest : int
            Approximate size in pixel of the smallest face that should be
            detected.
        size : int
            Size of the normalized face thumbnail.
        openface : str
            Path to openface FaceNet model.
        """
        super(Face, self).__init__()

        # face detection
        self._face_detector = dlib.get_frontal_face_detector()
        if smallest > SMALLEST_DEFAULT:
            self._upscale = 1
        else:
            self._upscale = int(np.ceil(SMALLEST_DEFAULT / smallest))

        # landmark detection
        if landmarks is not None:
            self._landmarks_detector = dlib.shape_predictor(landmarks)

        # normalization
        self.size = size
        self._landmarks = MINMAX_TEMPLATE * self.size

        if openface is not None:
            self._net = TorchWrap(model=openface, size=self.size, cuda=False)

    # face detection

    def iterfaces(self, rgb):
        """Iterate over all detected faces"""
        for face in self._face_detector(rgb, self._upscale):
            yield face

    # landmarks detection

    def _get_landmarks(self, rgb, face):
        """Return facial landmarks"""
        points = self._landmarks_detector(rgb, face).parts()
        landmarks = np.float32([(p.x, p.y) for p in points])
        return landmarks

    def landmarks(self, rgb, face):
        """Return facial landmarks"""
        return self._get_landmarks(rgb, face)

    # face normalization

    def _get_normalized(self, rgb, landmarks):
        """Return normalized face based on outer eyes and nose tip"""
        matrix = cv2.getAffineTransform(
            landmarks[OUTER_EYES_AND_NOSE],
            self._landmarks[OUTER_EYES_AND_NOSE])
        normalized = cv2.warpAffine(rgb, matrix, (self.size, self.size))
        return normalized

    def normalize(self, rgb, face):
        """Return normalized face"""
        landmarks = self._get_landmarks(rgb, face)
        return self._get_normalized(rgb, landmarks)

    # openface feature extraction

    def _get_openface(self, bgr):
        """(internal) openface feature extraction"""
        return self._net.forwardImage(bgr)

    def openface(self, rgb, face):
        """Return Openface features"""
        normalized_rgb = self.normalize(rgb, face)
        normalized_bgr = cv2.cvtColor(normalized_rgb, cv2.COLOR_BGR2RGB)
        return self._get_openface(normalized_bgr)

    # debugging

    def debug(self, image, face, landmarks):
        """Return face with overlaid landmarks"""
        copy = image.copy()
        for x, y in landmarks:
            cv2.rectangle(copy, (x, y), (x, y), (0, 255, 0), 2)
        copy = copy[face.top():face.bottom(),
                    face.left():face.right()]
        copy = cv2.resize(copy, (self.size, self.size))
        return copy

    def __call__(self, rgb, return_landmarks=False, return_normalized=False,
                 return_openface=False, return_debug=False):
        """Iterate over all faces

        Parameters
        ----------
        rgb : np.array
            RGB image to be processed
        return_landmarks : bool
            Whether to yield landmarks. Defaults to False.
        return_normalized : bool
            Whether to yield normalized face. Defaults to False.
        return_openface : bool
            Whether to yield openface descriptor. Defaults to False.
        return_debug : bool
            Whether to yield debugging image. Defaults to False.
        """

        for face in self.iterfaces(rgb):

            _face = (face.left(), face.top(), face.right(), face.bottom())

            # yield face if nothing else is asked for
            if not (return_landmarks or return_normalized or
                    return_openface or return_debug):
                yield _face
                continue

            # always return face as first tuple element
            result = (_face, )

            # compute landmarks
            landmarks = self._get_landmarks(rgb, face)

            # normalize face
            if return_normalized or return_openface:
                normalized = self._get_normalized(rgb, landmarks)

            # compute openface descriptor
            if return_openface:
                openface = self._get_openface(normalized)

            # compute debugging image
            if return_debug:
                debug = self.debug(rgb, face, landmarks)

            # append landmarks
            if return_landmarks:
                result = result + (landmarks, )

            # append normalized face
            if return_normalized:
                result = result + (normalized, )

            # append openface descriptor
            if return_openface:
                result = result + (openface, )

            # append debugging image
            if return_debug:
                result = result + (debug, )

            yield result



def simple_decorator(decorator):
    """This decorator can be used to turn simple functions
    into well-behaved decorators, so long as the decorators
    are fairly simple. If a decorator expects a function and
    returns a function (no descriptors), and if it doesn't
    modify function attributes or docstring, then it is
    eligible to use this. Simply apply @simple_decorator to
    your decorator and it will automatically preserve the
    docstring and function attributes of functions to which
    it is applied."""
    def new_decorator(f):
        g = decorator(f)
        g.__name__ = f.__name__
        g.__doc__ = f.__doc__
        g.__dict__.update(f.__dict__)
        return g

    # Now a few lines needed to make simple_decorator itself
    # be a well-behaved decorator.
    new_decorator.__name__ = decorator.__name__
    new_decorator.__doc__ = decorator.__doc__
    new_decorator.__dict__.update(decorator.__dict__)
    return new_decorator


@simple_decorator
def bufferize(iterator, with_context='center', context=1):
    def buffered_iterator(*args, **kwargs):

        timestamps = deque([], context)
        things = deque([], context)

        for timestamp, thing in iterator(*args, **kwargs):
            timestamps.append(timestamp)
            things.append(thing)
            if len(timestamps) < context:
                continue

            if with_context is 'center':
                yield timestamps[context / 2], things
            elif with_context is 'left':
                yield timestamp, things
            else:
                yield timestamps[0], things

        return buffered_iterator
